/*  GL Mixer - Mix multiple inputs to stereo output with OpenGL visualisation
    Copyright (C) 2003  Nick Lamb <njl+glmix@filter.tlrmx.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Some of this code is based on public domain examples from gtkglext */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <jack/jack.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkgl.h>

#include <GL/gl.h>
#include <GL/glu.h>

#define DEGREES (M_PI / 180.0)

/* Arbitrary non-zero constant */
#define BASE_PORT_NAME 64

#define TIMEOUT_FAST 40
#define TIMEOUT_SLOW 80
#define MOUSE_VELOCITY_FACTOR 600.0

/* Window size */
#define DEFAULT_WIDTH  640
#define DEFAULT_HEIGHT 400

/* Vertical Field Of View */
#define FOVY 60.0

#define WINDOW_TITLE "GL Mixer"

static gchar font_string[] = "Sans 12";
static GLuint font_list_base;
static gint font_height;

#define INPUT_PORTS 8
#define HITLIST_SIZE (INPUT_PORTS + 1) * 4

static char jack_client_name[] = "glmix";

typedef struct {
  jack_port_t *jack_port;
  float jack_level, draw_level;
  float volume, last_volume; /* 1.0 = 0dB FS */
  float pan, last_pan; /* 0.0 left 1.0 right */
  GLfloat color[4]; /* r, g, b, a */
  GLint x, y; /* screen co-ords of top */
  char name[64]; /* JACK port name */
  gchar alias[64]; /* on screen name */
} glmix_stream;

static glmix_stream streams[INPUT_PORTS] = {
 { 0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.5, {0.0, 0.0, 1.0, 0.4}, 0, 0, "input_1", "One" },
 { 0, 0.0, 0.0, 1.0, 0.0, 0.5, 0.5, {0.0, 0.5, 1.0, 0.4}, 0, 0, "input_2", "Two" },
 { 0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.5, {0.0, 1.0, 1.0, 0.4}, 0, 0, "input_3", "Three" },
 { 0, 0.0, 0.0, 0.5, 0.0, 1.0, 0.5, {0.0, 1.0, 0.5, 0.4}, 0, 0, "input_4", "Four" },
 { 0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.5, {0.0, 1.0, 0.0, 0.4}, 0, 0, "input_5", "Five" },
 { 0, 0.0, 0.0, 0.0, 0.0, 0.5, 0.5, {0.5, 1.0, 0.0, 0.4}, 0, 0, "input_6", "Six" },
 { 0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.5, {1.0, 1.0, 0.0, 0.4}, 0, 0, "input_7", "Seven" },
 { 0, 0.0, 0.0, 0.5, 0.0, 0.0, 0.5, {1.0, 0.5, 0.0, 0.4}, 0, 0, "input_8", "Eight" },
};

static jack_client_t *jack_handle;
static jack_port_t *left_port, *right_port;

static GLfloat actual_width, actual_height;
static GLuint draw_order[INPUT_PORTS]; /* arrray of indices into streams[] */
static const GLdouble zNear = 1.0, zFar = 99.1;
static GLfloat sphi = 0.0, stheta = -35.0;
static GLfloat zoom_distance = 12.3;
static int lastX, lastY;
static int stream = 0, selected = 0;

static guint source_status;
static guint32 redraw_rate = TIMEOUT_SLOW;
static gboolean spin = FALSE;

static void         timeout_speed     (GtkWidget   *widget, guint32 interval);
static void         timeout_none      (GtkWidget   *widget);

static void         show_status       (void);
static void         zoom              (gboolean    direction);
static void         toggle_spin       (GtkWidget   *widget);
static void         input_prefs ();
static void         output_prefs ();

static GdkGLConfig *configure_gl      (void);

static GtkWidget   *create_menu (GtkWidget   *drawing_area);
static GtkWidget   *create_window     (GdkGLConfig *glconfig);

static void font_setup (void);

static GtkWidget *main_window, *status_bar;

static float volume_curve(float volume)
{
  if (volume >= 1.0f) {
    return 1.0f;
  } else if (volume <= 0.0f) {
    return 0.0f;
  } else if (volume > 0.5f) {
    return powf(10.0, -2.0 + (volume * 2.0));
  } else {
    return powf(10.0, -5.0 + (volume * 8.0));
  }
}

static int playback_callback(jack_nframes_t nframes, void *arg)
{
    int k, frame;

    jack_default_audio_sample_t *left = (jack_default_audio_sample_t *)jack_port_get_buffer(left_port, nframes);
    jack_default_audio_sample_t *right = (jack_default_audio_sample_t *)jack_port_get_buffer(right_port, nframes);
    memset(left, 0, nframes * sizeof(jack_default_audio_sample_t));
    memset(right, 0, nframes * sizeof(jack_default_audio_sample_t));

    for (k= 0; k < INPUT_PORTS; ++k) {
      float pan = streams[k].last_pan * 0.5 * M_PI;
      float vol = streams[k].last_volume;
      const float vol_rate = (streams[k].volume - streams[k].last_volume)
                           / (float) nframes;
      const float pan_rate = (streams[k].pan - streams[k].last_pan) * M_PI
                           / (nframes * 2.0f);
      float level = streams[k].jack_level;
      jack_default_audio_sample_t *input = (jack_default_audio_sample_t *)jack_port_get_buffer(streams[k].jack_port, nframes);

      for (frame= 0; frame < nframes; ++frame) {
        pan += pan_rate;
        vol += vol_rate;
        right[frame] += input[frame] * sin(pan) * volume_curve(vol);
        left[frame] += input[frame] * cos(pan) * volume_curve(vol);
        if (fabsf(input[frame]) > level) level = fabsf(input[frame]);
      }

      streams[k].last_pan = streams[k].pan;
      streams[k].last_volume = streams[k].volume;
      streams[k].jack_level = level;
    }

    return(0);
}

static jack_client_t *open_jack(void)
{
    jack_client_t *jack_handle;
            
    if ((jack_handle = jack_client_new(jack_client_name)) == 0) {
        fprintf (stderr, "jack server not running ?\n");
        return NULL;
    }
    int k;
    for (k= 0; k < INPUT_PORTS; ++k) {
      streams[k].jack_port =
            jack_port_register(jack_handle, streams[k].name,
                               JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
    }
    left_port = jack_port_register(jack_handle, "left", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
    right_port = jack_port_register(jack_handle, "right", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
    jack_set_process_callback(jack_handle, playback_callback, 0);
    return(jack_handle);
}

/* -- ought to initialise OpenGL state here */

static void realize (GtkWidget *widget, gpointer data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return;

  font_setup();

  gdk_gl_drawable_gl_end (gldrawable);
}

static gboolean configure_event (GtkWidget *widget, GdkEventConfigure *event,
                                 gpointer data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  actual_width = widget->allocation.width;
  actual_height = widget->allocation.height;

  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return FALSE;

  /* make the GL view full the available area */
  glViewport (0, 0, actual_width, actual_height);

  gdk_gl_drawable_gl_end (gldrawable);

  return TRUE;
}

struct view_preset {
  GLfloat sphi;
  GLfloat stheta;
  GLfloat zoom_distance;
} preset_views[] = { { 0.0, -30.0, 12.3 } ,
                     { 180.0 , -45.0, 12.3 } ,
                     { -10.0 , -60.0,  15.0 } };

static void select_view(int view)
{
  sphi = preset_views[view].sphi;
  stheta = preset_views[view].stheta;
  zoom_distance = preset_views[view].zoom_distance;
}

/* How many squares to use in the grid, must be even */
static const float grid = 24.0f;

static void draw_grid(void)
{
  int k;
  GLfloat z = -0.5f;

  glBegin(GL_LINE_STRIP);
  for (k= 0; k <= grid; ++k) {
    glVertex3f( -0.5f + k / grid, 0.0, z);
    z = -z;
    glVertex3f( -0.5f + k / grid, 0.0, z);
  }
  for (k= 0; k <= grid; ++k) {
    glVertex3f( z, 0.0, 0.5f - k / grid);
    z = -z;
    glVertex3f( z, 0.0, 0.5f - k / grid);
  }
  glEnd();
}

static void draw_box(void)
{
  glBegin(GL_TRIANGLE_STRIP);
    glVertex3f(-0.5, 0.0, -0.5);
    glVertex3f(0.5, 0.0, -0.5);
    glVertex3f(-0.5, 1.0, -0.5);
    glVertex3f(0.5, 1.0, -0.5);
    glVertex3f(-0.5, 1.0, 0.5);
    glVertex3f(0.5, 1.0, 0.5);
    glVertex3f(-0.5, 0.0, 0.5);
    glVertex3f(0.5, 0.0, 0.5);
    glVertex3f(-0.5, 0.0, -0.5);
    glVertex3f(0.5, 0.0, -0.5);
  glEnd();

  glBegin(GL_TRIANGLE_STRIP);
    glVertex3f(-0.5, 0.0, -0.5);
    glVertex3f(-0.5, 0.0, 0.5);
    glVertex3f(-0.5, 1.0, -0.5);
    glVertex3f(-0.5, 1.0, 0.5);
  glEnd();
  glBegin(GL_TRIANGLE_STRIP);
    glVertex3f(0.5, 0.0, -0.5);
    glVertex3f(0.5, 0.0, 0.5);
    glVertex3f(0.5, 1.0, -0.5);
    glVertex3f(0.5, 1.0, 0.5);
  glEnd();

  /* outline */
  glBegin(GL_LINE_STRIP);
    glVertex3f(-0.5, 0.0, -0.5);
    glVertex3f(0.5, 0.0, -0.5);
    glVertex3f(0.5, 0.0, 0.5);
    glVertex3f(-0.5, 0.0, 0.5);
    glVertex3f(-0.5, 0.0, -0.5);
    glVertex3f(-0.5, 1.0, -0.5);
    glVertex3f(-0.5, 1.0, 0.5);
    glVertex3f(-0.5, 0.0, 0.5);
  glEnd();

  glBegin(GL_LINE_STRIP);
    glVertex3f(0.5, 1.0, 0.5);
    glVertex3f(-0.5, 1.0, 0.5);
    glVertex3f(-0.5, 1.0, -0.5);
    glVertex3f(0.5, 1.0, -0.5);
    glVertex3f(0.5, 1.0, 0.5);
    glVertex3f(0.5, 0.0, 0.5);
    glVertex3f(0.5, 0.0, -0.5);
    glVertex3f(0.5, 1.0, -0.5);
  glEnd();
}

static void draw_signal(int port)
{
  GLint viewport[4];
  GLdouble mvmatrix[16], projmatrix[16];
  GLdouble x, y, z;

  glPushMatrix();
  glTranslatef (streams[port].pan * 10.0, 0.0, 10.0 * streams[port].volume);

  glColor4fv(streams[port].color);
  glScalef(0.5, 2.2, 0.5);
  draw_box();

  glGetIntegerv (GL_VIEWPORT, viewport);
  glGetDoublev (GL_MODELVIEW_MATRIX, mvmatrix);
  glGetDoublev (GL_PROJECTION_MATRIX, projmatrix);
  gluProject(0.0, 1.2, 0.0, mvmatrix, projmatrix, viewport, &x, &y, &z);
  streams[port].x = (GLint) x;
  streams[port].y = (GLint) y;

  if (streams[port].draw_level > 0.0) {
    glColor4f(0.8, 0.0, 0.0, 1.0);
    glScalef(0.4, streams[port].draw_level, 0.4);
    draw_box();
  }
  glPopMatrix();
}

static void draw_scene(void)
{
  glInitNames();
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  glTranslatef (0.0, 0.0, -zoom_distance);
  glRotatef (-stheta, 1.0, 0.0, 0.0);
  glRotatef (sphi, 0.0, 1.0, 0.0);

  /* Draw grid */
  glPushName(0);
  glPushMatrix();
  glColor4f(0.7, 0.7, 0.7, 1.0);
  glScalef(10.5, 1.0, 10.5);
  draw_grid();
  glPopMatrix();

  /* Draw signal boxes */
  glTranslatef (-5.0, 0.0, -5.0);

  int k;
  for (k = 0; k < INPUT_PORTS; ++k) {
    const int port = draw_order[k];
    glLoadName(BASE_PORT_NAME + port);
    draw_signal(port);
  }
}

static void draw_labels(void)
{
  int j, k;

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluOrtho2D(0, actual_width, 0, actual_height);
  glDisable(GL_DEPTH_TEST);

  for (j = 0 ; j < INPUT_PORTS ; ++j) {
    const int len = strlen(streams[j].alias);
    if (j == stream) {
      glColor3f(1.0, 1.0, 1.0);
    } else {
      glColor3f(0.6, 0.7, 0.8);
    }
    glRasterPos2i (streams[j].x - (3.2 * len) , streams[j].y);
    for (k = 0; k < len; ++k) {
      const unsigned char c = (unsigned char) streams[j].alias[k];
      if (c > 32 && c < 127) {
        glCallList(c - 32 + font_list_base);
      } else {
        glCallList(font_list_base);
      }
    }
  }
}

static void order_objects(void)
{
  GLuint hitlist[HITLIST_SIZE];
  GLuint depth[INPUT_PORTS];
  GLint hits, k= 0, j= 0;

  for (k= 0; k < INPUT_PORTS; ++k) {
    draw_order[k]= k;
  }
  glSelectBuffer(HITLIST_SIZE, hitlist);
  glRenderMode(GL_SELECT);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (FOVY, (actual_width / actual_height), zNear, zFar);

  /* We're just figuring out the depth order this time */
  draw_scene();

  hits= glRenderMode(GL_RENDER);

  for (k= 0; k < hits; ++k) {
    const GLuint name = hitlist[4 * k + 3];
    if (name >= BASE_PORT_NAME)
      depth[name - BASE_PORT_NAME]= hitlist[4 * k + 1];
  }
  for (j= 0; j < INPUT_PORTS; ++j) {
    GLuint max_depth = 0;
    for (k= 0; k < INPUT_PORTS; ++k) {
      if (depth[k] > max_depth) {
        max_depth = depth[k];
        draw_order[j] = k;
      }
    }
    depth[draw_order[j]]= 0;
  }
}

static GLuint select_object(int x, int y)
{
  GLuint hitlist[HITLIST_SIZE];
  GLint hits, viewport[4];

  glSelectBuffer(HITLIST_SIZE, hitlist);
  glRenderMode(GL_SELECT);
  glGetIntegerv(GL_VIEWPORT, viewport);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPickMatrix(x, actual_height - y, 3, 3, viewport);
  gluPerspective (FOVY, (actual_width / actual_height), zNear, zFar);
  /* We're selecting the closest matching object */
  draw_scene();

  /* Now do the hard part */

  hits= glRenderMode(GL_RENDER);
  if (hits == 0)
    return 0;
  else
    /* hack */
    return hitlist[4 * (hits - 1) + 3];
}

static gboolean expose_event (GtkWidget *widget, GdkEventExpose *event,
                              gpointer data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return FALSE;

  order_objects();
  int k;
  for (k= 0; k < INPUT_PORTS; ++k) {
    float level= streams[k].jack_level;
    streams[k].jack_level = 0.0;
    if (level < streams[k].draw_level) level = streams[k].draw_level * 0.8 + level * 0.2;
    if (level <= 1.0e-20F) level = 0.0;
    streams[k].draw_level = level > 1.0 ? 1.0 : level;
  }

  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  glEnable(GL_LINE_SMOOTH);

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (FOVY, (actual_width / actual_height), zNear, zFar);

  /* This time we're actually drawing the scene */
  draw_scene();
  draw_labels();

  glFlush();

  /* Swap buffers */
  if (gdk_gl_drawable_is_double_buffered (gldrawable))
    gdk_gl_drawable_swap_buffers (gldrawable);
  else
    glFlush ();

  gdk_gl_drawable_gl_end (gldrawable);

  return TRUE;
}

static gboolean timeout (GtkWidget *widget)
{
  if (spin) {
    sphi+= 1.0;
  }
  gtk_widget_queue_draw (widget);

  return TRUE;
}

static void unrealize (GtkWidget *widget, gpointer data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return;

  gdk_gl_drawable_gl_end (gldrawable);
}

static gboolean motion_notify_event (GtkWidget *widget, GdkEventMotion *event,
                                     gpointer data)
{
/* actual_height is correct here, gluPerspective() uses vertical FOV */
  float xoff = (event->x - lastX) / actual_height;
  float yoff = (event->y - lastY) / actual_height;

  if (selected < BASE_PORT_NAME) {
    if (event->state & GDK_BUTTON1_MASK) {
      sphi += xoff * 80.0;
      stheta -= yoff * 40.0;
    }
  } else {
    if (event->state & GDK_BUTTON1_MASK) {
      float pan_off = cos(sphi * DEGREES) * xoff - sin(sphi * DEGREES) * yoff;
      float vol_off = sin(sphi * DEGREES) * xoff + cos(sphi * DEGREES) * yoff;

      streams[stream].pan += pan_off;
      if (streams[stream].pan < 0.0) streams[stream].pan = 0.0;
      if (streams[stream].pan > 1.0) streams[stream].pan = 1.0;
      streams[stream].volume += vol_off;
      if (streams[stream].volume < 0.0) streams[stream].volume = 0.0;
      if (streams[stream].volume > 1.0) streams[stream].volume = 1.0;
      show_status();
    }
  }

  lastX = event->x;
  lastY = event->y;

  return FALSE;
}

static gboolean scroll_event (GtkWidget *widget, GdkEventScroll *event,
                              gpointer data)
{
  if (event->direction == GDK_SCROLL_UP) {
    if (zoom_distance >= 5.0) zoom_distance -= 0.5;
    return TRUE;
  } else if (event->direction == GDK_SCROLL_DOWN) {
    if (zoom_distance <= 16.0) zoom_distance += 0.5;
    return TRUE;
  }

  return FALSE;
}

static gboolean button_press_event (GtkWidget *widget, GdkEventButton *event,
                                      gpointer data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  redraw_rate = TIMEOUT_FAST; /* while button is down */
  timeout_speed(widget, redraw_rate);
  gtk_widget_queue_draw (widget);

  if (event->button == 1) {
      lastX = event->x;
      lastY = event->y;
      gdk_gl_drawable_gl_begin (gldrawable, glcontext);
      selected = select_object(lastX, lastY);
      if (selected >= BASE_PORT_NAME) stream = selected - BASE_PORT_NAME;
      gdk_gl_drawable_gl_end (gldrawable);

      show_status();
      return TRUE;
  }

  if (event->button == 2) {
      lastX = event->x;
      lastY = event->y;
      return TRUE;
  }

  return FALSE;
}

static gboolean button_release_event (GtkWidget *widget, GdkEventButton *event,
                                    gpointer data)
{
  /* FIXME - what if press A, press B, release A ... */

  redraw_rate = spin ? TIMEOUT_FAST : TIMEOUT_SLOW;
  timeout_speed(widget, redraw_rate);
  gtk_widget_queue_draw (widget);

  return TRUE;
}

static gboolean key_press_event (GtkWidget *widget, GdkEventKey *event,
                                 gpointer data)
{
  switch (event->keyval) {

    case GDK_1:
      stream = 0;
      break;

    case GDK_2:
      stream = 1;
      break;

    case GDK_3:
      stream = 2;
      break;

    case GDK_4:
      stream = 3;
      break;

    case GDK_5:
      stream = 4;
      break;

    case GDK_6:
      stream = 5;
      break;

    case GDK_7:
      stream = 6;
      break;

    case GDK_8:
      stream = 7;
      break;

#if 0
    case GDK_9:
      stream = 8;
      break;

    case GDK_0:
      stream = 9;
      break;
#endif

    case GDK_Left:
      if (event->state & GDK_SHIFT_MASK) {
        streams[stream].pan = CLAMP(streams[stream].pan - 0.10, 0.0, 1.0);
      } else {
        streams[stream].pan = CLAMP(streams[stream].pan - 0.01, 0.0, 1.0);
      }
      break;

    case GDK_Right:
      if (event->state & GDK_SHIFT_MASK) {
        streams[stream].pan = CLAMP(streams[stream].pan + 0.10, 0.0, 1.0);
      } else {
        streams[stream].pan = CLAMP(streams[stream].pan + 0.01, 0.0, 1.0);
      }
      break;

    case GDK_Up:
      if (event->state & GDK_SHIFT_MASK) {
        streams[stream].volume = CLAMP(streams[stream].volume + 0.10, 0.0, 1.0);
      } else {
        streams[stream].volume = CLAMP(streams[stream].volume + 0.01, 0.0, 1.0);
      }
      break;

    case GDK_Down:
      if (event->state & GDK_SHIFT_MASK) {
        streams[stream].volume = CLAMP(streams[stream].volume - 0.10, 0.0, 1.0);
      } else {
        streams[stream].volume = CLAMP(streams[stream].volume - 0.01, 0.0, 1.0);
      }
      break;

    case GDK_ISO_Left_Tab: /* shift tab ?! */
      stream --;
      if (stream < 0) stream = INPUT_PORTS - 1;
      break;

    case GDK_Tab:
      stream ++;
      if (stream >= INPUT_PORTS) stream = 0;
      break;

    case GDK_s:
      toggle_spin (widget);
      break;

    case GDK_Escape:
      gtk_main_quit ();
      break;

    default:
      return FALSE;
  }

  show_status();

  return TRUE;
}

static guint timeout_id = 0;

static void
timeout_speed (GtkWidget *widget, guint32 interval)
{
  if (timeout_id != 0) {
    gtk_timeout_remove(timeout_id);
  }
  timeout_id = gtk_timeout_add (interval, (GtkFunction) timeout, widget);
}

static void
timeout_none (GtkWidget *widget)
{
  if (timeout_id != 0) {
    gtk_timeout_remove(timeout_id);
    timeout_id = 0;
  }
}

static gboolean map_event (GtkWidget *widget, GdkEvent *event, gpointer data)
{
  timeout_speed (widget, redraw_rate);

  return TRUE;
}

static gboolean unmap_event (GtkWidget *widget, GdkEvent *event,
                             gpointer data)
{
  timeout_none (widget);

  return TRUE;
}

static gboolean
visibility_notify_event (GtkWidget          *widget,
                         GdkEventVisibility *event,
                         gpointer            data)
{
  if (event->state == GDK_VISIBILITY_FULLY_OBSCURED) {
    timeout_none (widget);
  } else {
    timeout_speed (widget, redraw_rate);
  }

  return TRUE;
}


static void show_status (void)
{
  char text[256];
  float db = 10.0f * logf(volume_curve(streams[stream].volume));

  gtk_statusbar_pop(GTK_STATUSBAR(status_bar), source_status);
  if (streams[stream].pan < 0.495) {
    if (streams[stream].volume == 0.0)
      sprintf(text, "#%d %s - left, silent", stream+1, streams[stream].alias);
    else
      sprintf(text, "#%d %s - left, %3.0fdB", stream+1, streams[stream].alias, db);

  } else if (streams[stream].pan > 0.505) {
    if (streams[stream].volume == 0.0)
      sprintf(text, "#%d %s - right, silent", stream+1, streams[stream].alias);
    else
      sprintf(text, "#%d %s - right, %3.0fdB", stream+1, streams[stream].alias, db);
  } else {
    if (streams[stream].volume == 0.0)
      sprintf(text, "#%d %s - center, silent", stream+1, streams[stream].alias);
    else
      sprintf(text, "#%d %s - center, %3.0fdB", stream+1, streams[stream].alias, db);
  }
  gtk_statusbar_push(GTK_STATUSBAR(status_bar), source_status, text);
}


static void zoom (gboolean direction)
{
  if (zoom_distance >= 5.0 && direction == TRUE) {
    zoom_distance -= 1.0;
  } else if (zoom_distance <= 16.0 && direction == FALSE) {
    zoom_distance += 1.0;
  }
}


static void toggle_spin (GtkWidget *widget)
{
  spin = !spin;

  redraw_rate = spin ? TIMEOUT_FAST : redraw_rate;
  timeout_speed (widget, redraw_rate);
}


static GtkWidget *create_menu (GtkWidget *drawing_area)
{
  GtkWidget *menu = gtk_menu_bar_new();
  GtkWidget *submenu;
  GtkWidget *menu_item;

  /* File */
  submenu = gtk_menu_new ();
  menu_item = gtk_menu_item_new_with_mnemonic ("_File");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item), submenu);

  /* -> Quit */
  menu_item = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT, NULL);
  gtk_menu_shell_append (GTK_MENU_SHELL (submenu), menu_item);
  g_signal_connect (G_OBJECT (menu_item), "activate",
                    G_CALLBACK (gtk_main_quit), NULL);
        
  /* Audio */
  submenu = gtk_menu_new();
  menu_item = gtk_menu_item_new_with_mnemonic ("_Audio");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item) , submenu);

  /* -> Inputs */
  menu_item = gtk_menu_item_new_with_mnemonic ("_Inputs");
  gtk_menu_shell_append (GTK_MENU_SHELL (submenu), menu_item);
  g_signal_connect_swapped (G_OBJECT (menu_item), "activate",
                            G_CALLBACK (input_prefs), NULL);
  /* -> Outputs */
  menu_item = gtk_menu_item_new_with_mnemonic ("_Outputs");
  gtk_menu_shell_append (GTK_MENU_SHELL (submenu), menu_item);
  g_signal_connect_swapped (G_OBJECT (menu_item), "activate",
                            G_CALLBACK (output_prefs), NULL);

  /* View */
  submenu = gtk_menu_new ();
  menu_item = gtk_menu_item_new_with_mnemonic ("_View");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item) , submenu);

  /* Preset views */
  menu_item = gtk_menu_item_new_with_mnemonic("Front");
  gtk_menu_shell_append (GTK_MENU_SHELL (submenu), menu_item);
  g_signal_connect_swapped (G_OBJECT (menu_item), "activate",
                            G_CALLBACK (select_view), (gpointer) 0);
  menu_item = gtk_menu_item_new_with_mnemonic("Back");
  gtk_menu_shell_append (GTK_MENU_SHELL (submenu), menu_item);
  g_signal_connect_swapped (G_OBJECT (menu_item), "activate",
                            G_CALLBACK (select_view), (gpointer) 1);
  menu_item = gtk_menu_item_new_with_mnemonic("Top down");
  gtk_menu_shell_append (GTK_MENU_SHELL (submenu), menu_item);
  g_signal_connect_swapped (G_OBJECT (menu_item), "activate",
                            G_CALLBACK (select_view), (gpointer) 2);

  /* -> Zoom in /out */
  menu_item = gtk_menu_item_new_with_mnemonic("Zoom I_n");
  gtk_menu_shell_append (GTK_MENU_SHELL (submenu), menu_item);
  g_signal_connect_swapped (G_OBJECT (menu_item), "activate",
                            G_CALLBACK (zoom), (gpointer) TRUE);
  menu_item = gtk_menu_item_new_with_mnemonic("Zoom O_ut");
  gtk_menu_shell_append (GTK_MENU_SHELL (submenu), menu_item);
  g_signal_connect_swapped (G_OBJECT (menu_item), "activate",
                            G_CALLBACK (zoom), (gpointer) FALSE);

  /* -> Toggle spin */
  menu_item = gtk_check_menu_item_new_with_mnemonic("_Spin");
  gtk_menu_shell_append (GTK_MENU_SHELL (submenu), menu_item);
  g_signal_connect_swapped (G_OBJECT (menu_item), "toggled",
                            G_CALLBACK (toggle_spin), drawing_area);


  return menu;
}

static GtkWidget *create_window (GdkGLConfig *glconfig)
{
  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *drawing_area;
  GtkWidget *menu;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), WINDOW_TITLE);

  /* Get automatically redrawn if any of their children changed allocation. */
  gtk_container_set_reallocate_redraws (GTK_CONTAINER (window), TRUE);

  /* Connect signal handlers to the window */
  g_signal_connect (G_OBJECT (window), "delete_event",
                    G_CALLBACK (gtk_main_quit), NULL);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show (vbox);

  /* Drawing area to draw OpenGL scene */

  drawing_area = gtk_drawing_area_new ();
  gtk_widget_set_size_request (drawing_area, DEFAULT_WIDTH, DEFAULT_HEIGHT);

  /* Set OpenGL-capability to the widget */
  gtk_widget_set_gl_capability (drawing_area,
                                glconfig,
                                NULL,
                                TRUE,
                                GDK_GL_RGBA_TYPE);

  gtk_widget_add_events (drawing_area,
                         GDK_BUTTON1_MOTION_MASK    |
                         GDK_BUTTON2_MOTION_MASK    |
                         GDK_BUTTON3_MOTION_MASK    |
                         GDK_BUTTON_PRESS_MASK      |
                         GDK_BUTTON_RELEASE_MASK    |
                         GDK_SCROLL_MASK            |
                         GDK_VISIBILITY_NOTIFY_MASK);

  /* Connect signal handlers to the drawing area */
  g_signal_connect_after (G_OBJECT (drawing_area), "realize",
                          G_CALLBACK (realize), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "configure_event",
                    G_CALLBACK (configure_event), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "expose_event",
                    G_CALLBACK (expose_event), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "unrealize",
                    G_CALLBACK (unrealize), NULL);

  g_signal_connect (G_OBJECT (drawing_area), "motion_notify_event",
                    G_CALLBACK (motion_notify_event), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "button_press_event",
                    G_CALLBACK (button_press_event), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "scroll_event",
                    G_CALLBACK (scroll_event), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "button_release_event",
                    G_CALLBACK (button_release_event), NULL);

  /* key_press_event handler for top-level window */
  g_signal_connect_swapped (G_OBJECT (window), "key_press_event",
                            G_CALLBACK (key_press_event), drawing_area);

  /* Don't waste CPU when window isn't visible */
  g_signal_connect (G_OBJECT (drawing_area), "map_event",
                    G_CALLBACK (map_event), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "unmap_event",
                    G_CALLBACK (unmap_event), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "visibility_notify_event",
                    G_CALLBACK (visibility_notify_event), NULL);

  /* Menu Bar */
  menu = create_menu (drawing_area);

  gtk_box_pack_start (GTK_BOX (vbox), menu, FALSE, FALSE, 0);
  gtk_widget_show_all (menu);

  /* add the drawing area ... */
  gtk_box_pack_start (GTK_BOX (vbox), drawing_area, TRUE, TRUE, 0);
  gtk_widget_show (drawing_area);

  status_bar = gtk_statusbar_new();
  source_status = gtk_statusbar_get_context_id(GTK_STATUSBAR(status_bar), "source");

  gtk_box_pack_start (GTK_BOX (vbox), status_bar, FALSE, FALSE, 0);
  gtk_widget_show (status_bar);

  return window;
}

/* utility functions, more or less verbatim from gtkglext */

static GdkGLConfig *
configure_gl (void)
{
  GdkGLConfig *glconfig;

  /* Try double-buffered visual */
  glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB    |
                                        GDK_GL_MODE_DEPTH  |
                                        GDK_GL_MODE_DOUBLE);

  /* ... or a single-buffered equivalent */
  if (glconfig == NULL) {
    glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB | GDK_GL_MODE_DEPTH);
  }

  return glconfig;
}

static gboolean entry_defocus (GtkWidget *widget, GdkEventFocus *event,
                               gpointer user_data)
{
  int port = (int) user_data;
  const gchar *text = gtk_entry_get_text(GTK_ENTRY(widget));

  g_strlcpy(streams[port].alias, text, sizeof (streams[port].alias));

  return FALSE;
}

static void color_set (GtkColorButton *widget,
                       gpointer user_data)
{
  int port = (int) user_data;

  GdkColor color;
  guint16 alpha;

  alpha = gtk_color_button_get_alpha (widget);
  gtk_color_button_get_color (widget, &color);

  streams[port].color[0] = color.red / 65535.0;
  streams[port].color[1] = color.green / 65535.0;
  streams[port].color[2] = color.blue / 65535.0;
  streams[port].color[3] = alpha / 65535.0;
}

static GtkWidget * per_input_pref (int port)
{
  GdkColor color;
  GtkWidget *hbox = gtk_hbox_new(FALSE, 6);
  GtkWidget *entry = gtk_entry_new();
  GtkWidget *button;
  guint16 alpha = (int) (streams[port].color[3] * 65535.0);

  color.red = (int) (streams[port].color[0] * 65535.0);
  color.green = (int) (streams[port].color[1] * 65535.0);
  color.blue = (int) (streams[port].color[2] * 65535.0);
  
  button = gtk_color_button_new_with_color (&color);
  gtk_color_button_set_alpha ( GTK_COLOR_BUTTON(button), alpha);
  gtk_color_button_set_use_alpha  (GTK_COLOR_BUTTON (button), TRUE);

  gtk_entry_set_text(GTK_ENTRY (entry), streams[port].alias);
  gtk_box_pack_start(GTK_BOX( hbox), button, FALSE, FALSE, 6);
  gtk_box_pack_start(GTK_BOX( hbox), entry, TRUE, TRUE, 6);

  gtk_entry_set_activates_default ( GTK_ENTRY(entry), TRUE);
  g_signal_connect (G_OBJECT (entry), "focus-out-event",
                    G_CALLBACK (entry_defocus), (gpointer) port);
  g_signal_connect (G_OBJECT (button), "color-set",
                    G_CALLBACK (color_set), (gpointer) port);
  return hbox;
}

static void input_prefs ()
{
  GtkWidget *dialog =
      gtk_dialog_new_with_buttons("Inputs",
                                  GTK_WINDOW (main_window),
                                  GTK_DIALOG_DESTROY_WITH_PARENT,
                                  GTK_STOCK_CLOSE,
                                  GTK_RESPONSE_NONE,
                                  NULL);
  GtkWidget *label = gtk_label_new("Name");
  gtk_box_pack_start(GTK_BOX( GTK_DIALOG(dialog)->vbox), label, TRUE, TRUE, 6);

  int k;
  for (k = 0; k < INPUT_PORTS; ++k) {
    GtkWidget *hbox = per_input_pref (k);
    gtk_box_pack_start(GTK_BOX( GTK_DIALOG(dialog)->vbox), hbox, TRUE, TRUE, 6);
  }

  gtk_dialog_set_default_response (GTK_DIALOG(dialog), GTK_RESPONSE_NONE);

  g_signal_connect_swapped (dialog, "response",
                             G_CALLBACK (gtk_widget_destroy), dialog);

  gtk_widget_show_all(dialog);
}

void output_picked (GtkMenuItem *menuitem, gpointer user_data)
{
  jack_port_t *port = (jack_port_t *) user_data;
  const char *dest;

  dest= gtk_label_get_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(menuitem))));

  jack_port_disconnect(jack_handle, port);
  jack_connect(jack_handle, jack_port_name(port), dest);
}

static GtkWidget *per_output_pref (const char **ports, jack_port_t *port,
                                   const char *name)
{
  GtkWidget *hbox = gtk_hbox_new(FALSE, 6);
  const char **connected = jack_port_get_all_connections(jack_handle, port);

  GtkWidget *label = gtk_label_new(name);
  gtk_box_pack_start(GTK_BOX( hbox), label, FALSE, FALSE, 6);

  GtkWidget *options = gtk_option_menu_new();
  GtkWidget *menu = gtk_menu_new();
  const char **list = ports;
  int count = 0, picked = 0;
  GtkWidget *child = gtk_menu_item_new_with_label("(none)");
  g_signal_connect(child, "activate", G_CALLBACK(output_picked), (gpointer) port);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu), child);
  while (*list) {
    const char *portname = *list;
    count++;

    if (connected && !strcmp(portname, *connected)) {
      picked = count;
    }

    GtkWidget *child = gtk_menu_item_new_with_label(portname);
    g_signal_connect(child, "activate", G_CALLBACK(output_picked), (gpointer) port);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), child);

    list++;
  }
  gtk_option_menu_set_menu(GTK_OPTION_MENU(options), menu);
  gtk_option_menu_set_history(GTK_OPTION_MENU(options), picked);

  gtk_box_pack_start(GTK_BOX( hbox), options, TRUE, TRUE, 6);

  free(connected);
  return hbox;
}

static void output_prefs ()
{
  GtkWidget *dialog =
      gtk_dialog_new_with_buttons("Outputs",
                                  GTK_WINDOW (main_window),
                                  GTK_DIALOG_DESTROY_WITH_PARENT,
                                  GTK_STOCK_CLOSE,
                                  GTK_RESPONSE_NONE,
                                  NULL);

  const char **ports = jack_get_ports(jack_handle, NULL, NULL, JackPortIsInput);
  GtkWidget *label = gtk_label_new("Output from GL Mixer is a stereo pair");
  gtk_box_pack_start(GTK_BOX( GTK_DIALOG(dialog)->vbox), label, TRUE, TRUE, 6);

  GtkWidget *hbox;
  hbox = per_output_pref (ports, left_port, "Left");
  gtk_box_pack_start(GTK_BOX( GTK_DIALOG(dialog)->vbox), hbox, TRUE, TRUE, 6);
  hbox = per_output_pref (ports, right_port, "Right");
  gtk_box_pack_start(GTK_BOX( GTK_DIALOG(dialog)->vbox), hbox, TRUE, TRUE, 6);

  free(ports);

  gtk_dialog_set_default_response (GTK_DIALOG(dialog), GTK_RESPONSE_NONE);

  g_signal_connect_swapped (dialog, "response",
                             G_CALLBACK (gtk_widget_destroy), dialog);

  gtk_widget_show_all(dialog);
}


void font_setup (void)
{
  PangoFontDescription *font_desc;
  PangoFont *font;
  PangoFontMetrics *font_metrics;

  font_list_base = glGenLists (96);

  font_desc = pango_font_description_from_string (font_string);

  font = gdk_gl_font_use_pango_font (font_desc, 32, 96, font_list_base);

  if (font == NULL) return;

  font_metrics = pango_font_get_metrics (font, NULL);

  font_height = pango_font_metrics_get_ascent (font_metrics) +
                pango_font_metrics_get_descent (font_metrics);
  font_height = PANGO_PIXELS (font_height);

  pango_font_description_free (font_desc);
  pango_font_metrics_unref (font_metrics);
}

int main (int argc, char *argv[])
{
  GdkGLConfig *glconfig;
  GdkPixbuf *pixbuf;

  /* Initialize everything */
  gtk_init (&argc, &argv);
  gtk_gl_init (&argc, &argv);
  glconfig = configure_gl ();

  if (!glconfig) {
    GtkWidget *dialog = gtk_message_dialog_new (GTK_WINDOW(main_window),
                                                GTK_DIALOG_DESTROY_WITH_PARENT,
                                                GTK_MESSAGE_ERROR,
                                                GTK_BUTTONS_CLOSE,
        "OpenGL is misconfigured or unavailable.");
    gtk_dialog_run(GTK_DIALOG (dialog));
    gtk_widget_destroy (dialog);
    exit(1);
  }

  /* Create and show the application window. */
  main_window = create_window (glconfig);
  pixbuf = gdk_pixbuf_new_from_file ("glmixer.png", NULL);

  gtk_window_set_icon(GTK_WINDOW (main_window), pixbuf);
  gtk_widget_show (main_window);

  jack_handle = open_jack();
  if (!jack_handle || jack_activate(jack_handle)) {
    GtkWidget *dialog = gtk_message_dialog_new (GTK_WINDOW(main_window),
                                                GTK_DIALOG_DESTROY_WITH_PARENT,
                                                GTK_MESSAGE_ERROR,
                                                GTK_BUTTONS_CLOSE,
        "Cannot activate JACK. Start the JACK daemon before running GL Mixer.");
    gtk_dialog_run(GTK_DIALOG (dialog));
    gtk_widget_destroy (dialog);
    exit(1);
  }

  gtk_main ();

  jack_client_close(jack_handle);
  return 0;
}
